@@ -25,12 +25,15 @@
 #include "catalog/pg_am.h"
 #endif
 
+#include "catalog/pg_inherits_fn.h"
 #include "catalog/pg_namespace.h"
 #include "catalog/pg_opclass.h"
 #include "catalog/pg_type.h"
 #include "commands/tablecmds.h"
 #include "commands/trigger.h"
 #include "miscadmin.h"
+#include "storage/lmgr.h"
+#include "utils/array.h"
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
 #include "utils/rel.h"
@@ -61,6 +64,7 @@ extern Datum PGUT_EXPORT repack_swap(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT repack_drop(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT repack_disable_autovacuum(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT repack_index_swap(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT repack_get_table_and_inheritors(PG_FUNCTION_ARGS);
 
 PG_FUNCTION_INFO_V1(repack_version);
 PG_FUNCTION_INFO_V1(repack_trigger);
@@ -71,6 +75,7 @@ PG_FUNCTION_INFO_V1(repack_swap);
 PG_FUNCTION_INFO_V1(repack_drop);
 PG_FUNCTION_INFO_V1(repack_disable_autovacuum);
 PG_FUNCTION_INFO_V1(repack_index_swap);
+PG_FUNCTION_INFO_V1(repack_get_table_and_inheritors);
 
 static void	repack_init(void);
 static SPIPlanPtr repack_prepare(const char *src, int nargs, Oid *argtypes);
@@ -93,14 +98,6 @@ must_be_superuser(const char *func)
 }
 
 
-/* Include an implementation of RenameRelationInternal for old
- * versions which don't have one.
- */
-#if PG_VERSION_NUM < 80400
-static void RenameRelationInternal(Oid myrelid, const char *newrelname, Oid namespaceId);
-#endif
-
-
 /* The API of RenameRelationInternal() was changed in 9.2.
  * Use the RENAME_REL macro for compatibility across versions.
  */
@@ -151,7 +148,7 @@ repack_trigger(PG_FUNCTION_ARGS)
 
 	/* make sure it's called as a trigger at all */
 	if (!CALLED_AS_TRIGGER(fcinfo) ||
-		!TRIGGER_FIRED_BEFORE(trigdata->tg_event) ||
+		!TRIGGER_FIRED_AFTER(trigdata->tg_event) ||
 		!TRIGGER_FIRED_FOR_ROW(trigdata->tg_event) ||
 		trigdata->tg_trigger->tgnargs != 1)
 		elog(ERROR, "repack_trigger: invalid trigger call");
@@ -312,9 +309,9 @@ repack_apply(PG_FUNCTION_ARGS)
 			 * can delete all the rows we have processed at-once.
 			 */
 			if (i == 0)
-			    appendStringInfoString(&sql_pop, pkid);
+				appendStringInfoString(&sql_pop, pkid);
 			else
-			    appendStringInfo(&sql_pop, ",%s", pkid);
+				appendStringInfo(&sql_pop, ",%s", pkid);
 			pfree(pkid);
 		}
 		/* i must be > 0 (and hence we must have some rows to delete)
@@ -921,7 +918,7 @@ repack_swap(PG_FUNCTION_ARGS)
 	/* drop repack trigger */
 	execute_with_format(
 		SPI_OK_UTILITY,
-		"DROP TRIGGER IF EXISTS z_repack_trigger ON %s.%s CASCADE",
+		"DROP TRIGGER IF EXISTS repack_trigger ON %s.%s CASCADE",
 		nspname, relname);
 
 	SPI_finish();
@@ -962,7 +959,7 @@ repack_drop(PG_FUNCTION_ARGS)
 	 * To prevent concurrent lockers of the repack target table from causing
 	 * deadlocks, take an exclusive lock on it. Consider that the following
 	 * commands take exclusive lock on tables log_xxx and the target table
-	 * itself when deleting the z_repack_trigger on it, while concurrent
+	 * itself when deleting the repack_trigger on it, while concurrent
 	 * updaters require row exclusive lock on the target table and in
 	 * addition, on the log_xxx table, because of the trigger.
 	 *
@@ -1011,24 +1008,11 @@ repack_drop(PG_FUNCTION_ARGS)
 	{
 		execute_with_format(
 			SPI_OK_UTILITY,
-			"DROP TRIGGER IF EXISTS z_repack_trigger ON %s.%s CASCADE",
+			"DROP TRIGGER IF EXISTS repack_trigger ON %s.%s CASCADE",
 			nspname, relname);
 		--numobj;
 	}
 
-#if PG_VERSION_NUM < 80400
-	/* delete autovacuum settings */
-	execute_with_format(
-		SPI_OK_DELETE,
-		"DELETE FROM pg_catalog.pg_autovacuum v"
-		" USING pg_class c, pg_namespace n"
-		" WHERE relname IN ('log_%u', 'table_%u')"
-		"   AND n.nspname = 'repack'"
-		"   AND c.relnamespace = n.oid"
-		"   AND v.vacrelid = c.oid",
-		oid, oid);
-#endif
-
 	/* drop temp table */
 	if (numobj > 0)
 	{
@@ -1052,17 +1036,10 @@ repack_disable_autovacuum(PG_FUNCTION_ARGS)
 	/* connect to SPI manager */
 	repack_init();
 
-#if PG_VERSION_NUM >= 80400
 	execute_with_format(
 		SPI_OK_UTILITY,
 		"ALTER TABLE %s SET (autovacuum_enabled = off)",
 		get_relation_name(oid));
-#else
-	execute_with_format(
-		SPI_OK_INSERT,
-		"INSERT INTO pg_catalog.pg_autovacuum VALUES (%u, false, -1, -1, -1, -1, -1, -1, -1, -1)",
-		oid);
-#endif
 
 	SPI_finish();
 
@@ -1179,14 +1156,25 @@ swap_heap_or_index_files(Oid r1, Oid r2)
 		relform2->reltuples = swap_tuples;
 	}
 
+	indstate = CatalogOpenIndexes(relRelation);
+
+#if PG_VERSION_NUM < 100000
+
 	/* Update the tuples in pg_class */
 	simple_heap_update(relRelation, &reltup1->t_self, reltup1);
 	simple_heap_update(relRelation, &reltup2->t_self, reltup2);
 
 	/* Keep system catalogs current */
-	indstate = CatalogOpenIndexes(relRelation);
 	CatalogIndexInsert(indstate, reltup1);
 	CatalogIndexInsert(indstate, reltup2);
+
+#else
+
+	CatalogTupleUpdateWithInfo(relRelation, &reltup1->t_self, reltup1, indstate);
+	CatalogTupleUpdateWithInfo(relRelation, &reltup2->t_self, reltup2, indstate);
+
+#endif
+
 	CatalogCloseIndexes(indstate);
 
 	/*
@@ -1306,8 +1294,8 @@ repack_index_swap(PG_FUNCTION_ARGS)
 					 orig_idx_oid);
 	execute(SPI_OK_SELECT, str.data);
 	if (SPI_processed != 1)
-		elog(ERROR, "Could not find index 'index_%u', found %d matches",
-			 orig_idx_oid, SPI_processed);
+		elog(ERROR, "Could not find index 'index_%u', found " UINT64_FORMAT " matches",
+			 orig_idx_oid, (uint64) SPI_processed);
 
 	tuptable = SPI_tuptable;
 	desc = tuptable->tupdesc;
@@ -1318,27 +1306,52 @@ repack_index_swap(PG_FUNCTION_ARGS)
 	PG_RETURN_VOID();
 }
 
-#if PG_VERSION_NUM < 80400
-
-/* XXX: You might need to add PGDLLIMPORT into your miscadmin.h. */
-extern PGDLLIMPORT bool allowSystemTableMods;
-
-static void
-RenameRelationInternal(Oid myrelid, const char *newrelname, Oid namespaceId)
+/**
+ * @fn      Datum get_table_and_inheritors(PG_FUNCTION_ARGS)
+ * @brief   Return array containing Oids of parent table and its children.
+ *          Note that this function does not release relation locks.
+ *
+ * get_table_and_inheritors(table)
+ *
+ * @param	table	parent table.
+ * @retval	regclass[]
+ */
+Datum
+repack_get_table_and_inheritors(PG_FUNCTION_ARGS)
 {
-	bool	save_allowSystemTableMods = allowSystemTableMods;
+	Oid			parent = PG_GETARG_OID(0);
+	List	   *relations;
+	Datum	   *relations_array;
+	int			relations_array_size;
+	ArrayType  *result;
+	ListCell   *lc;
+	int			i;
 
-	allowSystemTableMods = true;
-	PG_TRY();
-	{
-		renamerel(myrelid, newrelname, OBJECT_TABLE);
-		allowSystemTableMods = save_allowSystemTableMods;
-	}
-	PG_CATCH();
-	{
-		allowSystemTableMods = save_allowSystemTableMods;
-		PG_RE_THROW();
-	}
-	PG_END_TRY();
+	LockRelationOid(parent, AccessShareLock);
+
+	/* Check that parent table exists */
+	if (!SearchSysCacheExists1(RELOID, ObjectIdGetDatum(parent)))
+		PG_RETURN_ARRAYTYPE_P(construct_empty_array(OIDOID));
+
+	/* Also check that children exist */
+	relations = find_all_inheritors(parent, AccessShareLock, NULL);
+
+	relations_array_size = list_length(relations);
+	if (relations_array_size == 0)
+		PG_RETURN_ARRAYTYPE_P(construct_empty_array(OIDOID));
+
+	relations_array = palloc(relations_array_size * sizeof(Datum));
+
+	i = 0;
+	foreach (lc, relations)
+		relations_array[i++] = ObjectIdGetDatum(lfirst_oid(lc));
+
+	result = construct_array(relations_array,
+							 relations_array_size,
+							 OIDOID, sizeof(Oid),
+							 true, 'i');
+
+	pfree(relations_array);
+
+	PG_RETURN_ARRAYTYPE_P(result);
 }
-#endif
